@@ -1,14 +1,15 @@
 """Main cloudwanderer module."""
 from typing import List
 import logging
-from typing import TYPE_CHECKING
+from typing import TYPE_CHECKING, Iterator
 import concurrent.futures
+from boto3.resources.base import ServiceResource
 from botocore import xform_name
 import boto3
 from .utils import exception_logging_wrapper
-from .boto3_interface import CloudWandererBoto3Interface, CustomAttributesInterface
+from .boto3_interface import CloudWandererBoto3Interface
 from .aws_urn import AwsUrn
-from .global_service_mappings import GlobalServiceMappingCollection
+from .service_mappings import ServiceMappingCollection, GlobalServiceResourceMappingNotFound
 from boto3.resources.model import ResourceModel
 
 logger = logging.getLogger('cloudwanderer')
@@ -32,13 +33,30 @@ class CloudWanderer():
         self.storage_connectors = storage_connectors
         self.boto3_session = boto3_session or boto3.session.Session()
         self.boto3_interface = CloudWandererBoto3Interface(boto3_session=self.boto3_session)
-        self.custom_attributes_interface = CustomAttributesInterface(boto3_session=self.boto3_session)
-        self.global_service_maps = GlobalServiceMappingCollection(boto3_session=self.boto3_session)
+        self.service_maps = ServiceMappingCollection(boto3_session=self.boto3_session)
         self._account_id = None
         self._enabled_regions = None
 
     def write_resources(
-            self, exclude_resources: List[str] = None, client_args: dict = None, concurrency: int = 1) -> None:
+            self, exclude_resources: List[str] = None, client_args: dict = None) -> None:
+        """Write all AWS resources in this account from all regions and all services to storage.
+
+        Arguments:
+            exclude_resources (list): A list of resource names to exclude (e.g. ``['instance']``)
+            client_args (dict): Arguments to pass into the boto3 client.
+                See: :meth:`boto3.session.Session.client`
+        """
+        logger.info('Writing resources in all regions')
+        for region_name in self.enabled_regions:
+            self.write_resources_in_region(
+                exclude_resources=exclude_resources,
+                region_name=region_name,
+                client_args=client_args
+            )
+
+    def write_resources_concurrently(
+            self, session_generator: callable, exclude_resources: List[str] = None, client_args: dict = None,
+            concurrency: int = 10) -> None:
         """Write all AWS resources in this account from all regions and all services to storage.
 
         Arguments:
@@ -50,15 +68,22 @@ class CloudWanderer():
                 multiple services to be queried concurrently in each region.
                 **WARNING:** Experimental. Complete data capture depends heavily on the thread safeness of the
                 storage connector and has not been thoroughly tested!
+            session_generator (callable): A method which returns a new boto3 session when called.
+                You're usually safe to pass in ``boto3.session.Session``
         """
         logger.info('Writing resources in all regions')
-        if concurrency > 1:
-            logger.warning('Using concurrency of: %s - CONCURRENCY IS EXPERIMENTAL', concurrency)
+        logger.warning('Using concurrency of: %s - CONCURRENCY IS EXPERIMENTAL', concurrency)
         with concurrent.futures.ThreadPoolExecutor(max_workers=concurrency) as executor:
             for region_name in self.enabled_regions:
+                boto3_session = session_generator() if session_generator else self.boto3_session
+                cw = CloudWanderer(
+                    storage_connectors=self.storage_connectors,
+                    boto3_session=boto3_session
+                )
+
                 executor.submit(
                     exception_logging_wrapper,
-                    method=self.write_resources_in_region,
+                    method=cw.write_resources_in_region,
                     exclude_resources=exclude_resources,
                     region_name=region_name,
                     client_args=client_args
@@ -108,12 +133,6 @@ class CloudWanderer():
 
         logger.info("Writing all %s resources in %s", service_name, client_args['region_name'])
         exclude_resources = exclude_resources or []
-        service_map = self.global_service_maps.get_global_service_map(service_name=service_name)
-        has_global_resources_in_this_region = service_map.has_global_resources_in_region(client_args['region_name'])
-        if not has_global_resources_in_this_region and not service_map.has_regional_resources:
-            logger.info("Skipping %s as it does not have resources in %s",
-                        service_name, client_args['region_name'])
-            return
 
         for resource_type in self.boto3_interface.get_service_resource_types(service_name=service_name):
             if resource_type in exclude_resources:
@@ -143,18 +162,30 @@ class CloudWanderer():
         client_args = client_args or {
             'region_name': region_name or self.boto3_session.region_name
         }
+        service_map = self.service_maps.get_service_mapping(service_name=service_name)
+        if service_map.is_global_service and service_map.global_service_region != client_args['region_name']:
+            logger.info("Skipping %s as it does not have resources in %s",
+                        service_name, client_args['region_name'])
+            return
         logger.info('--> Fetching %s %s from %s', service_name, resource_type, client_args['region_name'])
         resources = self.boto3_interface.get_resources_of_type(service_name, resource_type, client_args)
         urns = []
         for boto3_resource in resources:
-            urn = self._get_resource_urn(boto3_resource, client_args['region_name'])
-            if not self._should_write_resource_in_region(urn, client_args['region_name']):
-                continue
+            urns.extend(list(self._write_resource(boto3_resource, client_args['region_name'])))
+        self._clean_resources_in_region(service_name, resource_type, client_args['region_name'], urns)
+
+    def _write_resource(self, boto3_resource: ServiceResource, region_name: str) -> Iterator[AwsUrn]:
+        urn = self._get_resource_urn(boto3_resource, region_name)
+        for storage_connector in self.storage_connectors:
+            storage_connector.write_resource(urn, boto3_resource)
+        yield urn
+
+        for subresource in self.boto3_interface.get_subresources(boto3_resource=boto3_resource):
+            subresource.load()
+            urn = self._get_resource_urn(subresource, region_name)
+            yield urn
             for storage_connector in self.storage_connectors:
-                storage_connector.write_resource(urn, boto3_resource)
-            urns.append(urn)
-        self._clean_resources_in_region(
-            service_name, resource_type, client_args['region_name'], urns)
+                storage_connector.write_resource(urn, subresource)
 
     def _clean_resources_in_region(
             self, service_name: str, resource_type: str, region_name: str, current_urns: List[AwsUrn]) -> None:
@@ -168,19 +199,6 @@ class CloudWanderer():
                 urns_to_keep=current_urns
             )
 
-    def _should_write_resource_in_region(self, urn: AwsUrn, write_region: str) -> bool:
-        """Return True if this is a resource we should write in this region and log the result."""
-        if urn.region != write_region:
-            logger.debug(
-                "Skipping %s %s in %s because it is not in %s",
-                urn.resource_type,
-                urn.resource_id,
-                urn.region,
-                write_region
-            )
-            return False
-        return True
-
     def write_secondary_attributes(
             self, exclude_resources: List[str] = None, client_args: dict = None) -> None:
         """Write all secondary attributes in this account from all regions and all services to storage.
@@ -214,7 +232,7 @@ class CloudWanderer():
             client_args (dict): Arguments to pass into the boto3 client.
                 See: :meth:`boto3.session.Session.client`
         """
-        for boto3_service in self.custom_attributes_interface.get_all_resource_services():
+        for boto3_service in self.boto3_interface.get_all_resource_services():
             self.write_secondary_attributes_of_service_in_region(
                 service_name=boto3_service.meta.service_name,
                 exclude_resources=exclude_resources,
@@ -243,16 +261,28 @@ class CloudWanderer():
         client_args = client_args or {
             'region_name': region_name or self.boto3_session.region_name
         }
-        logger.info("Writing all %s secondary attributes in %s", service_name, client_args['region_name'])
         exclude_resources = exclude_resources or []
-        service_map = self.global_service_maps.get_global_service_map(service_name=service_name)
-        has_gobal_resources_in_this_region = service_map.has_global_resources_in_region(client_args['region_name'])
-        if not has_gobal_resources_in_this_region and not service_map.has_regional_resources:
+        service_map = self.service_maps.get_service_mapping(service_name=service_name)
+        if service_map.is_global_service and service_map.global_service_region != client_args['region_name']:
             logger.info("Skipping %s as it does not have resources in %s",
                         service_name, client_args['region_name'])
             return
         exclude_resources = exclude_resources or []
-        for resource_type in self.custom_attributes_interface.get_service_resource_types(service_name):
+        collections = self.boto3_interface.get_resource_collections(
+            boto3_service=self.boto3_interface.get_resource_service_by_name(
+                service_name=service_name,
+                client_args=client_args
+            )
+        )
+        for resource_type in self.boto3_interface.get_service_resource_types_from_collections(collections):
+            try:
+                resource_map = service_map.get_resource_mapping(resource_type=resource_type)
+            except GlobalServiceResourceMappingNotFound:
+                continue
+            if not resource_map.has_secondary_attributes:
+                continue
+            logger.info("Writing all %s %s secondary attributes in %s",
+                        service_name, resource_type, client_args['region_name'])
             if resource_type in exclude_resources:
                 logger.info('Skipping %s as per exclude_resources', resource_type)
                 continue
@@ -280,20 +310,34 @@ class CloudWanderer():
         client_args = client_args or {
             'region_name': region_name or self.boto3_session.region_name
         }
-        logger.info('--> Fetching %s %s in %s', service_name, resource_type, client_args['region_name'])
-        secondary_attributes = self.custom_attributes_interface.get_resources_of_type(
+        resources = self.boto3_interface.get_resources_of_type(
             service_name=service_name,
             resource_type=resource_type,
             client_args=client_args
         )
-        for secondary_attribute in secondary_attributes:
-            urn = self._get_resource_urn(secondary_attribute, client_args['region_name'])
-            for storage_connector in self.storage_connectors:
-                storage_connector.write_secondary_attribute(
-                    urn=urn,
-                    secondary_attribute=secondary_attribute,
-                    attribute_type=xform_name(secondary_attribute.meta.resource_model.name)
-                )
+        service_map = self.service_maps.get_service_mapping(service_name=service_name)
+        try:
+            resource_map = service_map.get_resource_mapping(resource_type=resource_type)
+        except GlobalServiceResourceMappingNotFound:
+            logger.warning('Skipping %s %s as it does not have secondary attributes', service_name, resource_type)
+            return
+        if not resource_map.has_secondary_attributes:
+            logger.warning('Skipping %s %s as it does not have secondary attributes', service_name, resource_type)
+            return
+        for resource in resources:
+            secondary_attributes = self.boto3_interface.get_secondary_attributes(
+                boto3_resource=resource)
+            for secondary_attribute in secondary_attributes:
+                attribute_type = xform_name(secondary_attribute.meta.resource_model.name)
+                logger.info('--> Fetching %s %s %s in %s', service_name,
+                            resource_type, attribute_type, client_args['region_name'])
+                urn = self._get_resource_urn(resource, client_args['region_name'])
+                for storage_connector in self.storage_connectors:
+                    storage_connector.write_secondary_attribute(
+                        urn=urn,
+                        secondary_attribute=secondary_attribute,
+                        attribute_type=attribute_type
+                    )
 
     @property
     def account_id(self) -> str:
@@ -316,15 +360,19 @@ class CloudWanderer():
         return self._enabled_regions
 
     def _get_resource_urn(self, resource: ResourceModel, region_name: str) -> 'AwsUrn':
-        id_member_name = resource.meta.resource_model.identifiers[0].name
-        resource_id = getattr(resource, id_member_name)
-        if resource_id.startswith('arn:'):
-            resource_id = ''.join(resource_id.split(':')[5:])
-        global_service_map = self.global_service_maps.get_global_service_map(resource.meta.service_name)
+        id_members = [x.name for x in resource.meta.resource_model.identifiers]
+        resource_ids = []
+        for id_member in id_members:
+            id_part = getattr(resource, id_member)
+            if id_part.startswith('arn:'):
+                id_part = ''.join(id_part.split(':')[5:])
+            resource_ids.append(id_part)
+        compound_resource_id = ':'.join(resource_ids)
+        service_map = self.service_maps.get_service_mapping(resource.meta.service_name)
         return AwsUrn(
             account_id=self.account_id,
-            region=global_service_map.get_resource_region(resource, region_name),
+            region=service_map.get_resource_region(resource, region_name),
             service=resource.meta.service_name,
-            resource_type=xform_name(resource.meta.resource_model.shape),
-            resource_id=resource_id
+            resource_type=xform_name(resource.meta.resource_model.name),
+            resource_id=compound_resource_id
         )
